/* ************************************************************************* */
/*                                                                           */
/*    hk.c: Dual timezone clock for the TS7400                               */
/*                                                                           */
/*    Copyright (C) 2010  Len Bayles, TMD Innovations                        */
/*                                                                           */
/*    This program is free software: you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation, either version 3 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program.  If not, see <http://www.gnu.org/licenses.    */
/*                                                                           */
/* ************************************************************************* */

/* Version 0.2.0 */

#include <fcntl.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>


#define	CONTROL_WAIT	250000


static	char	qcode[16] = {'X','R','L','X','L','X','X','R','R','X','X','L','X','L','R','X'};


#define	CLOCK			0x10
#define DATA			0x20
#define DISP0                   0x40
#define DISP1                   0x80

#define	MODE_LOCAL		1
#define	MODE_DUAL		2
#define	MODE_UTC		3

#define	DEFAULT_INTENSITY	0x00

#define	HK_DISPLAY_FIFO		"/tmp/hk_display"
#define	HK_MODE_FIFO		"/tmp/hk_mode"


int	daemonize();
int	cloop();
int	local(int fd, long now);
int	dual(int fd, long now);
int	utc(int fd, long now);
int	display(unsigned char *gpioregs);
int	dwrite(unsigned char *gpioregs, char *buffer);
int	bwrite(unsigned char *gpioregs, char byte);
int	latch(unsigned char *gpioregs, char device);
int	dimmer(unsigned char *gpioregs);
int	mode(unsigned char *gpioregs);
int	init_fifo();
int	init_display();
int	init_io(unsigned char **gpioregs);


int	main()
	{
	int		fd, status;
	unsigned char	*gpioregs;


	daemonize();

	if((fd = init_io(&gpioregs)) == -1)
		return(-1);

	if(0 != (status = init_fifo()))
		return(status);

	status = fork();		/* Fork off Display Writer */

	if(status < 0)
		exit(127);
	else if(status == 0)
		{
		display(gpioregs);
		exit(10);
		}

	if(init_display())		/* Init display drivers */
		return(-1);


	status = fork();		/* Fork off Mode Switch routine */

	if(status < 0)
		exit(128);
	else if(status == 0)
		{
		mode(gpioregs);
		exit(2);
		}


	status = fork();		/* Fork off Dimmer control routine */

	if(status < 0)
		exit(129);
	else if(status == 0)
		{
		dimmer(gpioregs);
		exit(3);
		}

	cloop();

	return(0);
	}


int	daemonize()
	{
	pid_t		pid, sid;


	pid = fork();

	if(pid < 0)
		exit(1);

	if(pid > 0)
		exit(0);

	umask(0);

	sid = setsid();

	if(sid < 0)
		exit(2);

	if((chdir("/")) < 0)
		exit(3);

	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	return(0);
	}


int	cloop()
	{
	int			fd, mfd, mode = MODE_DUAL;
	char			buffer;
	long			now, past = 0;
	struct	timespec	req, rem;


	req.tv_sec = 0;
	req.tv_nsec = 1000;

	if((mfd = open(HK_MODE_FIFO, O_RDONLY | O_NONBLOCK)) < 0)
		return(-1);

	if((fd = open(HK_DISPLAY_FIFO, O_WRONLY)) > 0)
		{
		while(1)
			{
			time(&now);

			if(1 == read(mfd, &buffer, 1))
				{
				switch(buffer)
					{
					case 'R':
						if(mode < 3)
							++mode;
						break;
					case 'L':
						if(mode > 1)
							--mode;
						break;
					default:
						break;

					}
				--past;
				}

			if(past != now)
				{
				switch(mode)
					{
					case MODE_DUAL:
						dual(fd, now);
						break;
					case MODE_LOCAL:
						local(fd, now);
						break;
					case MODE_UTC:
						utc(fd, now);
						break;
					default:
						mode = MODE_DUAL;
						break;
					}

				past = now;
				}

			nanosleep(&req, &rem);
			}
		}

	return(0);
	}


int	dual(int fd, long now)
	{
	unsigned char	buffer[3];
	struct tm	*gm, *local;


	gm = gmtime(&now);

	buffer[0] = DISP0;			/* Tens of hours */
	buffer[1] = 0x01;
	buffer[2] = gm->tm_hour/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Hours */
	buffer[1] = 0x02;
	buffer[2] = (gm->tm_hour % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Tens of Minutes */
	buffer[1] = 0x03;
	buffer[2] = (gm->tm_min/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Minutes */
	buffer[1] = 0x04;
	buffer[2] = (gm->tm_min % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Tens of seconds */
	buffer[1] = 0x05;
	buffer[2] = (gm->tm_sec/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Seconds */
	buffer[1] = 0x06;
	buffer[2] = gm->tm_sec % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);


	local = localtime(&now);

	buffer[0] = DISP1;			/* Tens of hours */
	buffer[1] = 0x01;
	buffer[2] = local->tm_hour/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Hours */
	buffer[1] = 0x02;
	buffer[2] = (local->tm_hour % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Tens of Minutes */
	buffer[1] = 0x03;
	buffer[2] = (local->tm_min/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Minutes */
	buffer[1] = 0x04;
	buffer[2] = (local->tm_min % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Tens of seconds */
	buffer[1] = 0x05;
	buffer[2] = (local->tm_sec/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Seconds */
	buffer[1] = 0x06;
	buffer[2] = local->tm_sec % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	return(0);
	}


int	local(int fd, long now)
	{
	unsigned char	buffer[3];
	struct tm	*local;

	
	local = localtime(&now);

	local->tm_mon += 1;

	buffer[0] = DISP0;			/* Tens of Months */
	buffer[1] = 0x01;
	buffer[2] = local->tm_mon/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Months */
	buffer[1] = 0x02;
	buffer[2] = local->tm_mon % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Tens of Days */
	buffer[1] = 0x03;
	buffer[2] = local->tm_mday/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Days */
	buffer[1] = 0x04;
	buffer[2] = local->tm_mday % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Tens of Years */
	buffer[1] = 0x05;
	buffer[2] = (local->tm_year/10) % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Years */
	buffer[1] = 0x06;
	buffer[2] = local->tm_year % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);


	buffer[0] = DISP1;			/* Tens of hours */
	buffer[1] = 0x01;
	buffer[2] = local->tm_hour/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Hours */
	buffer[1] = 0x02;
	buffer[2] = (local->tm_hour % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Tens of Minutes */
	buffer[1] = 0x03;
	buffer[2] = (local->tm_min/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Minutes */
	buffer[1] = 0x04;
	buffer[2] = (local->tm_min % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Tens of seconds */
	buffer[1] = 0x05;
	buffer[2] = (local->tm_sec/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Seconds */
	buffer[1] = 0x06;
	buffer[2] = local->tm_sec % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);




	return(0);
	}


int	utc(int fd, long now)
	{
	unsigned char	buffer[3];
	struct tm	*gm;

	
	gm = gmtime(&now);

	gm->tm_mon += 1;

	buffer[0] = DISP0;			/* Tens of Months */
	buffer[1] = 0x01;
	buffer[2] = gm->tm_mon/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Months */
	buffer[1] = 0x02;
	buffer[2] = gm->tm_mon % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Tens of Days */
	buffer[1] = 0x03;
	buffer[2] = gm->tm_mday/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Days */
	buffer[1] = 0x04;
	buffer[2] = gm->tm_mday % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Tens of Years */
	buffer[1] = 0x05;
	buffer[2] = (gm->tm_year/10) % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP0;			/* Years */
	buffer[1] = 0x06;
	buffer[2] = gm->tm_year % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);


	buffer[0] = DISP1;			/* Tens of hours */
	buffer[1] = 0x01;
	buffer[2] = gm->tm_hour/10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Hours */
	buffer[1] = 0x02;
	buffer[2] = (gm->tm_hour % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Tens of Minutes */
	buffer[1] = 0x03;
	buffer[2] = (gm->tm_min/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Minutes */
	buffer[1] = 0x04;
	buffer[2] = (gm->tm_min % 10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Tens of seconds */
	buffer[1] = 0x05;
	buffer[2] = (gm->tm_sec/10) | 0x80;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	buffer[0] = DISP1;			/* Seconds */
	buffer[1] = 0x06;
	buffer[2] = gm->tm_sec % 10;

	if(3 != write(fd, &buffer, 3))
		return(-1);

	return(0);
	}


int	display(unsigned char *gpioregs)
	{
	int			fd;
	unsigned char		buffer[3];
	struct	timespec	req, rem;


	req.tv_sec = 0;
	req.tv_nsec = 1000;

	if((fd = open(HK_DISPLAY_FIFO, O_RDONLY)) > 0)
		{
		while(1)
			{
			if(3 == read(fd, &buffer, 3))
				dwrite(gpioregs, buffer);
			else
				nanosleep(&req, &rem);
			}

		close(fd);
		}
	else
		return(-1);

	return(0);
	}


int	dwrite(unsigned char *gpioregs, char *buffer)
	{
	bwrite(gpioregs, buffer[1]);
	bwrite(gpioregs, buffer[2]);
	latch(gpioregs, buffer[0]);

	return(0);
	}

/* WORK */

int	bwrite(unsigned char *gpioregs, char byte)
	{
	int	i;


	for(i = 8 ; i ; i--)
		{
		if(byte & 0x80)
			gpioregs[2] |= DATA;
		else
			gpioregs[2] &= ~DATA;

		gpioregs[2] |= CLOCK; 
		gpioregs[2] &= !CLOCK; 
                byte <<= 1;
		}

	return(0);
	}


int	latch(unsigned char *gpioregs, char device)
	{
        gpioregs[2] |= device;
	gpioregs[2] &= ~device;

	return(0);
	}


int	mode(unsigned char *gpioregs)
	{
	int			change = 0, fd;
	char			buffer;
	unsigned char		last, new;
        struct  timespec        req, rem;


	while((fd = open(HK_MODE_FIFO, O_WRONLY | O_NONBLOCK)) < 0)
		;

        req.tv_sec = 0;
        req.tv_nsec = 1000;

	last = (gpioregs[2] >> 2) & 0x03;

	while(1)
		{
		if(last != (new = ((gpioregs[2] >> 2) & 0x03)))
			{
/* printf("%c %X %X\n",qcode[(last<< 2) | new], last, new); */

			buffer = qcode[(last<< 2) | new];

			if(1 != write(fd, &buffer, 1))
				return(-1);

			last = new;

			change = CONTROL_WAIT;
			}
		else
			{
			if(change)
				--change;
			else
				nanosleep(&req, &rem);
			}
		}


	return(0);
	}


int	dimmer(unsigned char *gpioregs)
	{
	int			change = 0, fd;
	unsigned char		buffer[3], last, new;
	unsigned char		intensity = DEFAULT_INTENSITY;
        struct  timespec        req, rem;


        req.tv_sec = 0;
        req.tv_nsec = 1000;

	last = gpioregs[2] & 0x03;

	if((fd = open(HK_DISPLAY_FIFO, O_WRONLY)) < 0)
		return(-1);

	while(1)
		{
		if(last != (new = (gpioregs[2] & 0x03)))
			{
		/* printf("%c %X %X\n",qcode[(last<< 2) | new], last, new); */

			switch(qcode[(last<< 2) | new])
				{
				case 'R':
					if(intensity < 15)
						++intensity;

					/* Set Intensity DISP0 */

					buffer[0] = DISP0;	
					buffer[1] = 0x0a;
					buffer[2] = intensity;

					if(3 != write(fd, &buffer, 3))
						return(-1);

					/* Set Intensity DISP1 */

					buffer[0] = DISP1;	

					if(3 != write(fd, &buffer, 3))
						return(-1);

					break;
				case 'L':
					if(intensity)
						--intensity;

					/* Set Intensity DISP0 */

					buffer[0] = DISP0;	
					buffer[1] = 0x0a;
					buffer[2] = intensity;

					if(3 != write(fd, &buffer, 3))
						return(-1);

					/* Set Intensity DISP1 */

					buffer[0] = DISP1;	

					if(3 != write(fd, &buffer, 3))
						return(-1);

					break;
				default:
					break;
				}

			last = new;

			change = CONTROL_WAIT;
			}
		else
			{
			if(change)
				--change;
			else
				nanosleep(&req, &rem);
			}
		}

	close(fd);

	return(0);
	}


int	init_fifo()
	{
	struct	stat	buf;


	if(stat(HK_DISPLAY_FIFO, &buf) < 0)	/* Create Display FIFO */
		{
		if(mkfifo(HK_DISPLAY_FIFO, S_IRUSR | S_IWUSR) < 0)
			return(1);	/* Create DFIFO Failed */
		}
	else
		{
		if(!S_ISFIFO(buf.st_mode))
			return(2);	/* Filename in use */
		}


	if(stat(HK_MODE_FIFO, &buf))	/* Create Mode FIFO */
		{
		if(mkfifo(HK_MODE_FIFO, S_IRUSR | S_IWUSR) < 0)
			return(3);	/* Create MFIFO Failed */
		}
	else
		{
		if(!S_ISFIFO(buf.st_mode))
			return(4);	/* Filename in use */
		}

	return(0);
	}


int	init_display()
	{
	int		fd;
	unsigned char	buffer[3];


	if((fd = open(HK_DISPLAY_FIFO, O_WRONLY)) > 0)
		{
		buffer[0] = DISP0;	/* Test mode DISP0 */
		buffer[1] = 0x0f;
		buffer[2] = 0xff;

		if(3 != write(fd, &buffer, 3))
			return(-1);

		buffer[0] = DISP1;	/* Test mode DISP1 */
		buffer[1] = 0x0f;
		buffer[2] = 0xff;

		if(3 != write(fd, &buffer, 3))
			return(-1);


	        sleep(1);


		buffer[0] = DISP0;	/* Test mode off DISP0 */
		buffer[1] = 0x0f;
		buffer[2] = 0x00;

		if(3 != write(fd, &buffer, 3))
			return(-1);

		buffer[0] = DISP1;	/* Test mode off DISP1 */
		buffer[1] = 0x0f;
		buffer[2] = 0x00;

		if(3 != write(fd, &buffer, 3))
			return(-1);


		buffer[0] = DISP0;	/* Set Normal mode DISP0 */
		buffer[1] = 0x0c;
		buffer[2] = 0x01;

		if(3 != write(fd, &buffer, 3))
			return(-1);

		buffer[0] = DISP1;	/* Set Normal mode DISP1 */
		buffer[1] = 0x0c;
		buffer[2] = 0x01;

		if(3 != write(fd, &buffer, 3))
			return(-1);


		buffer[0] = DISP0;	/* Set decode mode DISP0 */
		buffer[1] = 0x09;
		buffer[2] = 0xff;

		if(3 != write(fd, &buffer, 3))
			return(-1);

		buffer[0] = DISP1;	/* Set decode mode DISP1 */
		buffer[1] = 0x09;
		buffer[2] = 0xff;

		if(3 != write(fd, &buffer, 3))
			return(-1);


		buffer[0] = DISP0;	/* Set Intensity DISP0 */
		buffer[1] = 0x0a;
		buffer[2] = DEFAULT_INTENSITY;

		if(3 != write(fd, &buffer, 3))
			return(-1);

		buffer[0] = DISP1;	/* Set Intensity DISP1 */
		buffer[1] = 0x0a;
		buffer[2] = DEFAULT_INTENSITY;

		if(3 != write(fd, &buffer, 3))
			return(-1);


		buffer[0] = DISP0;	/* Set Scan Limit DISP0 */
		buffer[1] = 0x0b;
		buffer[2] = 0x05;

		if(3 != write(fd, &buffer, 3))
			return(-1);

		buffer[0] = DISP1;	/* Set Scan Limit DISP1 */
		buffer[1] = 0x0b;
		buffer[2] = 0x05;

		if(3 != write(fd, &buffer, 3))
			return(-1);

		close(fd);
		}
	else
		return(-1);

	return(0);
	}


/* Fix this function to be kinder...  Don't make assumptions for IO not used. */

int	init_io(unsigned char **gpioregs)
	{
	int	fd;


	if((fd = open("/dev/mem", O_RDWR | O_SYNC)) < 0)
		return(-1);

	*gpioregs = (char *)mmap(0, 4096, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0x12c00000);

	(*gpioregs)[0] = 0xF0; /* dir reg 0 -> 3 inputs ; 4 -> 7 outputs */

	(*gpioregs)[2] = 0x0F; 

	return(fd);
	}


